class Episode():
    def __init__(self, id:str, name:str, episode_no:str, url:str, m3u8:str) -> None:
        self.id = id
        self.name = name
        self.episode_no = episode_no
        self.url = url
        self.m3u8 = m3u8


import base64
from bs4 import BeautifulSoup
import re
import requests

from ..object.episode import Episode
from .strategy import Strategy


EPISODE_PATTERN = '^https:\/\/www\.ktkkt\.cc\/play\/([\d|-]*)\.html$'
SERIES_PATTERN = '^https:\/\/www\.ktkkt\.cc\/movie\/index(\d+)\.html$'
SERIES_URL = 'https://www.ktkkt.cc/movie/index{series_id}.html'
# EPISODE_URL = 'https://www.ktkkt.cc/play/{episode_id}.html'
EPISODE_URL = 'https://www.ktkkt.cc{episode_href}'


class KtkktStrategy(Strategy):
    def __init__(self) -> None:
        super().__init__()

    def get_series(self, url:str) -> list:
        result = re.search(SERIES_PATTERN, url, re.IGNORECASE)
        if result is not None:
            req = requests.get(url)
            doc = BeautifulSoup(req.content, 'html.parser')
            name = doc.select_one('h1.title.text-fff').text.strip()
            elements = doc.select('a[target="_self"][title]')
            return list(map(
                lambda e: Episode(
                    e.attrs['href'].split('/')[-1].replace('.html', '').strip(),
                    name,
                    e.attrs['title'].strip(),
                    EPISODE_URL.format(episode_href=e.attrs['href']),
                    None
                ),
                elements
            ))
        else:
            raise ValueError(url)

    def get_episode(self, url:str) -> Episode:
        result = re.search(EPISODE_PATTERN, url, re.IGNORECASE)
        if result is not None:
            req = requests.get(url)
            doc = BeautifulSoup(req.content, 'html.parser')
            return Episode(
                result.group(1).strip(),
                doc.select_one('h3.title.text-fff').text.strip(),
                doc.select_one('#playlist a[style="color:red"]').attrs['title'].strip(),
                url,
                base64.b64decode(req.text.split('now=base64decode("')[1].split('");')[0]).decode("utf-8")
            )
        else:
            raise ValueError(url)

    def get_m3u8(self, url:str) -> str:
        req = requests.get(url)
        encoded = req.text.split('now=base64decode("')[1].split('");')[0]
        return base64.b64decode(encoded).decode("utf-8")





from io import StringIO
import unittest
from unittest.mock import patch
from unittest import TestCase

from ktkkt_downloader.logger import Logger
logger = Logger(__name__)


MESSAGE = 'TEST_MESSAGE'


class TestLogger(TestCase):
    def test_print(self):
        with patch('sys.stdout', new = StringIO()) as fake_out:
            print(MESSAGE)
            self.assertEqual(fake_out.getvalue(), '{}\n'.format(MESSAGE))

    def test_info(self):
        with TestCase.assertLogs(__name__, level='INFO') as cm:
            logger.info(MESSAGE)
            self.assertEqual(cm.output, ['INFO:{}:{}'.format(__name__, MESSAGE)])

    def test_warning(self):
        with TestCase.assertLogs(__name__, level='WARN') as cm:
            logger.warning(MESSAGE)
            self.assertEqual(cm.output, ['WARNING:{}:{}'.format(__name__, MESSAGE)])

    def test_error(self):
        with TestCase.assertLogs(__name__, level='ERROR') as cm:
            logger.error(MESSAGE)
            self.assertEqual(cm.output, ['ERROR:{}:{}'.format(__name__, MESSAGE)])


if __name__ == '__main__':
    unittest.main()


from ktkkt_downloader.downloader import Downloader

from .scraper import KtkktStrategy, Scraper

from .logger import Logger
logger = Logger(__name__)


FILENAME = '{name}.E{episode_no}.mp4'


class KtkktDownloader():
    def __init__(self, dir:str) -> None:
        self.downloader = Downloader(dir)
        self.scraper = Scraper()
        self.scraper.set_strategy(KtkktStrategy())

    def download_series(self, url:str):
        episodes = self.scraper.get_series(url)
        for episode in episodes:
            if episode.m3u8 is None:
                episode.m3u8 = self.scraper.get_m3u8(episode.url)
            self.downloader.m3u8(
                episode.m3u8,
                FILENAME.format(
                    name=episode.name,
                    episode_no=episode.episode_no
                )
            )

    def download_episode(self, url:str):
        episode = self.scraper.get_episode(url)
        if episode.m3u8 is None:
            episode.m3u8 = self.scraper.get_m3u8(episode.url)
        self.downloader.m3u8(
            episode.m3u8,
            FILENAME.format(
                name=episode.name,
                episode_no=episode.episode_no
            )
        )



import unittest
from unittest import TestCase

from ktkkt_downloader.object import Episode
from ktkkt_downloader.scraper import KtkktStrategy, Scraper


class TestScraper(TestCase):
    def test_ktkkt_series(self):
        ktkktStrategy = KtkktStrategy()
        scraper = Scraper()
        scraper.set_strategy(ktkktStrategy)
        episodes = scraper.get_series('https://www.ktkkt.cc/movie/index8080.html')
        self.assertEqual(episodes[0].id, '8080-0-0')
        self.assertEqual(episodes[0].name, '龙珠Z粤语')
        self.assertEqual(episodes[0].episode_no, '01')
        self.assertEqual(episodes[0].url, 'https://www.ktkkt.cc/play/8080-0-0.html')

    def test_ktkkt_episode(self):
        ktkktStrategy = KtkktStrategy()
        scraper = Scraper()
        scraper.set_strategy(ktkktStrategy)
        episode = scraper.get_episode('https://www.ktkkt.cc/play/8080-0-0.html')
        self.assertEqual(episode.id, '8080-0-0')
        self.assertEqual(episode.name, '龙珠Z粤语')
        self.assertEqual(episode.episode_no, '01')
        self.assertEqual(episode.url, 'https://www.ktkkt.cc/play/8080-0-0.html')


if __name__ == '__main__':
    unittest.main()


import ffmpeg
import os
from pathvalidate import sanitize_filename
import traceback

from ..logger import Logger
logger = Logger(__name__)


class Downloader:
    def __init__(self, dir:str) -> None:
        self.dir = dir
    
    def m3u8(self, m3u8:str, filename:str) -> None:
        try:
            output = os.path.join(
                self.dir,
                sanitize_filename(filename)
            )
            os.makedirs(os.path.dirname(output), exist_ok=True)
            if os.path.exists(output) is False:
                stream = ffmpeg.input(m3u8)
                stream = ffmpeg.output(stream, output, vcodec='copy', acodec='copy')
                ffmpeg.run(stream)
                logger.info('Downloaded {} -> {}'.format(m3u8, os.path.abspath(output)))
            else:
                raise FileExistsError(output)
        except:
            logger.error(traceback.format_exc())




import logging


logging.basicConfig(
    level=logging.INFO, 
    format='%(asctime)s-%(name)s-%(levelname)-8s: %(message)s',
)


class Logger():
    def __init__(self, logger_name:str) -> None:
        self.logger = logging.getLogger(logger_name)

    def info(self, *messages) -> None:
        for message in messages: self.logger.info(message)

    def warning(self, *messages) -> None:
        for message in messages: self.logger.warning(message)

    def error(self, *messages) -> None:
        for message in messages: self.logger.error(message)



from abc import ABC, abstractmethod

class Strategy():
    def __init__(self) -> None:
        pass
    
    @abstractmethod
    def get_series(self, url:str):
        return
    
    @abstractmethod
    def get_episode(self, url:str):
        return
    
    @abstractmethod
    def get_m3u8(self, url:str):
        return
    


from abc import ABC, abstractmethod
from typing import Union

from ..object import Episode
from .strategy import Strategy

# context
class Scraper():
    def set_strategy(self, strategy:Strategy):
        self.strategy = strategy

    def get_series(self, url:str) -> list:
        return self.strategy.get_series(url)
    
    def get_episode(self, url:str) -> Episode:
        return self.strategy.get_episode(url)
    
    def get_m3u8(self, url:str) -> str:
        return self.strategy.get_m3u8(url)